/*	-- $Header$ --
 *
 *	projekt   :  CAN - Controller Area Network
 *
 *	purpose   :  CANopen-over-TCP/IP Client (helper)
 *
 *	copyright :  (c) 2009, UV Software, Friedrichshafen
 *
 *	compiler  :  GCC - GNU C Compiler (Linux Kernel 2.6)
 *
 *	export    :  (see header file)
 *
 *	includes  :  tcp_open.h
 *
 *	author    :  Uwe Vogt, UV Software, Friedrichshafen
 *
 *	e-mail    :  uwe.vogt@uv-software.de
 *
 *
 *	-----------  description  -----------------------------------------------
 *
 *	CANopen-over-TCP/IP Client - Helper Functions
 *
 *
 *	-----------  history  ---------------------------------------------------
 *
 *	$Log$
 */


/*  -----------  includes  -------------------------------------------------
 */

#include "tcp_open.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "base64.h"

/*  -----------  defines  --------------------------------------------------
 */

#define WHITESPACE(x)		(' ' == (x) || (x) == '\t')
#define VISIBLE(x)			(' ' <= (x) && (x) <= '\177')
#define SIGN(x)				('+' == (x) || (x) == '-')
#define DECIMAL(x)			('0' <= (x) && (x) <= '9')
#define OCTAL(x)			('0' <= (x) && (x) <= '7')
#define HEXADECIMAL(x)		('0' <= (x) && (x) <= '9')||('a' <= (x) && (x) <= 'f')||('A' <= (x) && (x) <= 'F')
#define BASE64(x)			('0' <= (x) && (x) <= '9')||('a' <= (x) && (x) <= 'z')||('A' <= (x) && (x) <= 'Z')||((x) == '+')||((x) == '/')

#define ERROR_OFFSET		600


/*  - - - - - -  Keywords  - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

#define BOOLEAN				0
#define COMMUNICATION		1
#define COPYRIGHT			2
#define DISABLE				3
#define DOMAIN				4
#define EMCY				5
#define ENABLE				6
#define ERREUR				7/*@ToDo: FIXME!*/
#define EVENT				8
#define EXECUTE				9
#define EXIT				10
#define FIRMWARE			11
#define GUARDING			12
#define HARDWARE			13
#define HEARTBEAT			14
#define ID					15
#define INFO				16
#define INIT				17
#define INTEGER8			18
#define INTEGER16			19
#define INTEGER24			20
#define INTEGER32			21
#define INTEGER40			22
#define INTEGER48			23
#define INTEGER56			24
#define INTEGER64			25
#define NETWORK				26
#define NODE				27
#define OCTET_STRING		28
#define OKAY				29/*@ToDo: FIXME!*/
#define PDO					30
#define PREOPERATIONAL		31
#define READ				32
#define REAL32				33
#define REAL64				34
#define RECEIVE				35
#define RESET				36
#define RESTORE				37
#define RPDO				38
#define RTR					39
#define SDO_TIME_OUT		40
#define SEND				41
#define SET					42
#define SOFTWARE			43
#define START				44
#define STOP				45
#define STORE				46
#define SYNC				47
#define TIME_OF_DAY			48
#define TIME_DIFFERENCE		49
#define TPDO				50
#define UNICODE_STRING		51
#define UNSIGNED8			52
#define UNSIGNED16			53
#define UNSIGNED24			54
#define UNSIGNED32			55
#define UNSIGNED40			56
#define UNSIGNED48			57
#define UNSIGNED56			58
#define UNSIGNED64			59
#define USER				60
#define VERSION				61
#define VISIBLE_STRING		62
#define WAIT				63
#define WRITE				64

/*  -----------  types  ----------------------------------------------------
 */


/*  -----------  prototypes  -----------------------------------------------
 */

static int lookahead(char *line, int *pos);
static int character(char *line, int *pos);
static int token(char *line, int *pos);
static int compare(char *string, char *keyword);

static int ascii2unsigned8(char *line, int *pos, unsigned char *value);
static int ascii2unsigned16(char *line, int *pos, unsigned short *value);
static int ascii2unsigned32(char *line, int *pos, unsigned long *value);
static int ascii2integer8(char *line, int *pos, char *value);
static int ascii2integer16(char *line, int *pos, short *value);
static int ascii2integer32(char *line, int *pos, long *value);

static int ascii2domain(char *line, int *pos, unsigned char *buffer, int *length, int nbyte);
static int ascii2string(char *line, int *pos, char *buffer, int *length, int nbyte);

/*  -----------  variables  ------------------------------------------------
 */


/*  -----------  functions  ------------------------------------------------
 */

int tcp_eval_response(char* response)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	switch(token(response, &pos)) {
	case ERREUR:
		/* token ERROR read: */
		if((chr = lookahead(response, &pos)) == -1)
			return COPERR_FORMAT;
		if((chr = character(response, &pos)) != ':')
			return COPERR_FORMAT;
		/* scan the <error-code> */
		if((chr = lookahead(response, &pos)) == -1)
			return COPERR_FORMAT;
		if(!ascii2integer32(response, &pos, &error_code))
			return COPERR_FORMAT;
		return (ERROR_OFFSET + error_code) * (-1);
	case OKAY:
		/* token OK read: */
		return TCPERR_NoError;
	default:
		return COPERR_FORMAT;
	}
}

int tcp_eval_sdo_response(char* response, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	switch(token(response, &pos)) {
	case ERREUR:
		/* token ERROR read: */
		if((chr = lookahead(response, &pos)) == -1)
			return COPERR_FORMAT;
		if((chr = character(response, &pos)) != ':')
			return COPERR_FORMAT;
		/* scan the <error-code> */
		if((chr = lookahead(response, &pos)) == -1)
			return COPERR_FORMAT;
		if(!ascii2integer32(response, &pos, &error_code))
			return COPERR_FORMAT;
		if(error_code >= 0x05000000) {
			if(abort_code)
			  *abort_code = error_code;
			return TCPERR_Aborted;
		}
		return (ERROR_OFFSET + error_code) * (-1);
	case OKAY:
		/* token OK read: */
		return TCPERR_NoError;
	default:
		return COPERR_FORMAT;
	}
}

int tcp_eval_sdo_unsigned8(char* response, BYTE* value, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(DECIMAL(chr)) {
		if(!value)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2unsigned8(response, &pos, value))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_unsigned16(char* response, WORD* value, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(DECIMAL(chr)) {
		if(!value)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2unsigned16(response, &pos, value))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_unsigned32(char* response, DWORD* value, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(DECIMAL(chr)) {
		if(!value)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2unsigned32(response, &pos, value))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_integer8(char* response, CHAR* value, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(SIGN(chr) || DECIMAL(chr)) {
		if(!value)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2integer8(response, &pos, value))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_integer16(char* response, SHORT* value, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(SIGN(chr) || DECIMAL(chr)) {
		if(!value)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2integer16(response, &pos, value))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_integer32(char* response, LONG* value, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(SIGN(chr) || DECIMAL(chr)) {
		if(!value)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2integer32(response, &pos, value))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_domain(char* response, short* length, BYTE* data, short nbyte, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!compare(&response[pos], "Error:")) {
		if(!length || !data)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2domain(response, &pos, data, (int*)length, (int)nbyte))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_sdo_string(char* response, short* length, char* data, short nbyte, long* abort_code)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!compare(&response[pos], "Error:")) {
		if(!length || !data)
			return COPERR_NULLPTR;
		if(abort_code)
		  *abort_code = 0x0;
		/* scan the <value> */
		if(!ascii2string(response, &pos, data, (int*)length, (int)nbyte))
			return COPERR_FORMAT;
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			if(error_code >= 0x05000000) {
				if(abort_code)
				  *abort_code = error_code;
				return TCPERR_Aborted;
			}
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_can_message(char* response, WORD* cob_id, short* length, BYTE* data)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;
	short i;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(DECIMAL(chr)) {
		if(!cob_id || !length || !data)
			return COPERR_NULLPTR;
		/* scan the <cob-id> */
		if(!ascii2unsigned16(response, &pos, cob_id))
			return COPERR_FORMAT;
		/* scan the <length> */
		if(!ascii2integer16(response, &pos, length))
			return COPERR_FORMAT;
		/* scan the {<value>}* */
		for(i = 0; i < *length && i < 8; i++) {
			if(!ascii2unsigned8(response, &pos, &data[i]))
				return COPERR_FORMAT;
		}
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			return (ERROR_OFFSET + error_code) * (-1);
		case OKAY:
			/* token OK read: */
			return COPERR_RX_EMPTY;
		default:
			return COPERR_FORMAT;
		}
	}
}

int tcp_eval_rtr_message(char* response, short* length, BYTE* data)
{
	unsigned long sequence = 0;
	long error_code = 0;
	int pos = 0, chr;
	short i;

	/* scan the literal '[' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != '[')
		return COPERR_FORMAT;
	/* scan the <sequence> */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(!ascii2unsigned32(response, &pos, &sequence))
		return COPERR_FORMAT;
	/* scan the literal ']' */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if((chr = character(response, &pos)) != ']')
		return COPERR_FORMAT;
	/* scan the answer */
	if((chr = lookahead(response, &pos)) == -1)
		return COPERR_FORMAT;
	if(DECIMAL(chr)) {
		if(!length || (*length && !data))
			return COPERR_NULLPTR;
		/* scan the <length> */
		if(!ascii2integer16(response, &pos, length))
			return COPERR_FORMAT;
		/* scan the {<value>}* */
		for(i = 0; i < *length && i < 8; i++) {
			if(!ascii2unsigned8(response, &pos, &data[i]))
				return COPERR_FORMAT;
		}
		return TCPERR_NoError;
	}
	else {
		switch(token(response, &pos)) {
		case ERREUR:
			/* token ERROR read: */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if((chr = character(response, &pos)) != ':')
				return COPERR_FORMAT;
			/* scan the <error-code> */
			if((chr = lookahead(response, &pos)) == -1)
				return COPERR_FORMAT;
			if(!ascii2integer32(response, &pos, &error_code))
				return COPERR_FORMAT;
			return (ERROR_OFFSET + error_code) * (-1);
		default:
			return COPERR_FORMAT;
		}
	}
}

/*  -----------  local functions  ------------------------------------------
 */

static int lookahead(char *line, int *pos)
{
	if(!line || !pos)
		return -1;
	while(line[*pos]) {
		if(!WHITESPACE(line[*pos]))
			return line[*pos];
		else
			*pos += 1;
	}
	return -1;
}

static int character(char *line, int *pos)
{
	int chr;

	if(!line || !pos)
		return -1;
	if((chr = line[*pos]))
		*pos += 1;
	else
		chr = -1;
	return chr;
}

static int token(char *line, int *pos)
{
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(compare(&line[*pos], "b")) {
		*pos += strlen("b");
		return BOOLEAN;
	}
	if(compare(&line[*pos], "communication")) {
		*pos += strlen("communication");
		return COMMUNICATION;
	}
	if(compare(&line[*pos], "comm")) {
		*pos += strlen("comm");
		return COMMUNICATION;
	}
	if(compare(&line[*pos], "copyright")) {
		*pos += strlen("copyright");
		return COPYRIGHT;
	}
	if(compare(&line[*pos], "disable")) {
		*pos += strlen("disable");
		return DISABLE;
	}
	if(compare(&line[*pos], "d")) {
		*pos += strlen("d");
		return DOMAIN;
	}
	if(compare(&line[*pos], "emcy")) {
		*pos += strlen("emcy");
		return EMCY;
	}
	if(compare(&line[*pos], "enable")) {
		*pos += strlen("enable");
		return ENABLE;
	}
	if(compare(&line[*pos], "error")) {
		*pos += strlen("error");
		return ERREUR;
	}
	if(compare(&line[*pos], "event")) {
		*pos += strlen("event");
		return EVENT;
	}
	if(compare(&line[*pos], "execute")) {
		*pos += strlen("execute");
		return EXECUTE;
	}
	if(compare(&line[*pos], "exec")) {
		*pos += strlen("exec");
		return EXECUTE;
	}
	if(compare(&line[*pos], "exit")) {
		*pos += strlen("exit");
		return EXIT;
	}
	if(compare(&line[*pos], "e")) {
		*pos += strlen("e");
		return ERREUR;
	}
	if(compare(&line[*pos], "firmware")) {
		*pos += strlen("firmware");
		return FIRMWARE;
	}
	if(compare(&line[*pos], "guarding")) {
		*pos += strlen("guarding");
		return GUARDING;
	}
	if(compare(&line[*pos], "hardware")) {
		*pos += strlen("hardware");
		return HARDWARE;
	}
	if(compare(&line[*pos], "heartbeat")) {
		*pos += strlen("heartbeat");
		return HEARTBEAT;
	}
	if(compare(&line[*pos], "i8")) {
		*pos += strlen("i8");
		return INTEGER8;
	}
	if(compare(&line[*pos], "i16")) {
		*pos += strlen("i16");
		return INTEGER16;
	}
	if(compare(&line[*pos], "i24")) {
		*pos += strlen("i24");
		return INTEGER24;
	}
	if(compare(&line[*pos], "i32")) {
		*pos += strlen("i32");
		return INTEGER32;
	}
	if(compare(&line[*pos], "i40")) {
		*pos += strlen("i40");
		return INTEGER40;
	}
	if(compare(&line[*pos], "i48")) {
		*pos += strlen("i48");
		return INTEGER48;
	}
	if(compare(&line[*pos], "i56")) {
		*pos += strlen("i56");
		return INTEGER56;
	}
	if(compare(&line[*pos], "i64")) {
		*pos += strlen("i64");
		return INTEGER64;
	}
	if(compare(&line[*pos], "id")) {
		*pos += strlen("id");
		return ID;
	}
	if(compare(&line[*pos], "info")) {
		*pos += strlen("info");
		return INFO;
	}
	if(compare(&line[*pos], "init")) {
		*pos += strlen("init");
		return INIT;
	}
	if(compare(&line[*pos], "network")) {
		*pos += strlen("network");
		return NETWORK;
	}
	if(compare(&line[*pos], "node")) {
		*pos += strlen("node");
		return NODE;
	}
	if(compare(&line[*pos], "ok")) {
		*pos += strlen("ok");
		return OKAY;
	}
	if(compare(&line[*pos], "os")) {
		*pos += strlen("os");
		return OCTET_STRING;
	}
	if(compare(&line[*pos], "pdo")) {
		*pos += strlen("pdo");
		return PDO;
	}
	if(compare(&line[*pos], "preoperational")) {
		*pos += strlen("preoperational");
		return PREOPERATIONAL;
	}
	if(compare(&line[*pos], "preop")) {
		*pos += strlen("preop");
		return PREOPERATIONAL;
	}
	if(compare(&line[*pos], "p")) {
		*pos += strlen("p");
		return PDO;
	}
	if(compare(&line[*pos], "r32")) {
		*pos += strlen("r32");
		return REAL32;
	}
	if(compare(&line[*pos], "r64")) {
		*pos += strlen("r64");
		return REAL64;
	}
	if(compare(&line[*pos], "read")) {
		*pos += strlen("read");
		return READ;
	}
	if(compare(&line[*pos], "receive")) {
		*pos += strlen("receive");
		return RECEIVE;
	}
	if(compare(&line[*pos], "recv")) {
		*pos += strlen("recv");
		return RECEIVE;
	}
	if(compare(&line[*pos], "restore")) {
		*pos += strlen("restore");
		return RESTORE;
	}
	if(compare(&line[*pos], "reset")) {
		*pos += strlen("reset");
		return RESET;
	}
	if(compare(&line[*pos], "rpdo")) {
		*pos += strlen("rpdo");
		return RPDO;
	}
	if(compare(&line[*pos], "rtr")) {
		*pos += strlen("rtr");
		return RTR;
	}
	if(compare(&line[*pos], "r")) {
		*pos += strlen("r");
		return READ;
	}
	if(compare(&line[*pos], "sdo_timeout")) {
		*pos += strlen("sdo_timeout");
		return SDO_TIME_OUT;
	}
	if(compare(&line[*pos], "set")) {
		*pos += strlen("set");
		return SET;
	}
	if(compare(&line[*pos], "send")) {
		*pos += strlen("send");
		return SEND;
	}
	if(compare(&line[*pos], "software")) {
		*pos += strlen("software");
		return SOFTWARE;
	}
	if(compare(&line[*pos], "start")) {
		*pos += strlen("start");
		return START;
	}
	if(compare(&line[*pos], "stop")) {
		*pos += strlen("stop");
		return STOP;
	}
	if(compare(&line[*pos], "store")) {
		*pos += strlen("store");
		return STORE;
	}
	if(compare(&line[*pos], "sync")) {
		*pos += strlen("sync");
		return SYNC;
	}
	if(compare(&line[*pos], "td")) {
		*pos += strlen("td");
		return TIME_DIFFERENCE;
	}
	if(compare(&line[*pos], "tpdo")) {
		*pos += strlen("tpdo");
		return TPDO;
	}
	if(compare(&line[*pos], "t")) {
		*pos += strlen("t");
		return TIME_OF_DAY;
	}
	if(compare(&line[*pos], "u8")) {
		*pos += strlen("u8");
		return UNSIGNED8;
	}
	if(compare(&line[*pos], "u16")) {
		*pos += strlen("u16");
		return UNSIGNED16;
	}
	if(compare(&line[*pos], "u24")) {
		*pos += strlen("u24");
		return UNSIGNED24;
	}
	if(compare(&line[*pos], "u32")) {
		*pos += strlen("u32");
		return UNSIGNED32;
	}
	if(compare(&line[*pos], "u40")) {
		*pos += strlen("u40");
		return UNSIGNED40;
	}
	if(compare(&line[*pos], "u48")) {
		*pos += strlen("u48");
		return UNSIGNED48;
	}
	if(compare(&line[*pos], "u56")) {
		*pos += strlen("u56");
		return UNSIGNED56;
	}
	if(compare(&line[*pos], "u64")) {
		*pos += strlen("u64");
		return UNSIGNED64;
	}
	if(compare(&line[*pos], "user")) {
		*pos += strlen("user");
		return USER;
	}
	if(compare(&line[*pos], "us")) {
		*pos += strlen("us");
		return UNICODE_STRING;
	}
	if(compare(&line[*pos], "version")) {
		*pos += strlen("version");
		return VERSION;
	}
	if(compare(&line[*pos], "vs")) {
		*pos += strlen("vs");
		return VISIBLE_STRING;
	}
	if(compare(&line[*pos], "wait")) {
		*pos += strlen("wait");
		return WAIT;
	}
	if(compare(&line[*pos], "write")) {
		*pos += strlen("write");
		return WRITE;
	}
	if(compare(&line[*pos], "w")) {
		*pos += strlen("w");
		return WRITE;
	}
	return -1;
}

static int compare(char *string, char *keyword)
{
	char *p1, *p2;

	if(!string || !keyword)
		return 0;
	for(p1 = string, p2 = keyword; *p2 != '\0'; p1++, p2++) {
		if(*p1 == '\0')
			return 0;
		if(tolower(*p1) != tolower(*p2))
			return 0;
	}
	return 1;
}

static int ascii2unsigned8(char *line, int *pos, unsigned char *value)
{
	unsigned char num = 0;
	int n = 0;

	if(value)
		*value = 0;
	if(!line || !pos)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '0') {
		num = (unsigned char)(line[*pos] - '0');
		*pos += 1;
		if(line[*pos] == 'x' || line[*pos] == 'X') {
			*pos += 1;
			/* hexadecimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; HEXADECIMAL(line[*pos]); *pos += 1, n++) {
				if(DECIMAL(line[*pos]))
					num = (num << 4) + (unsigned char)(line[*pos] - '0');
				else if('a' <= line[*pos] && line[*pos] <= 'f')
					num = (num << 4) + (unsigned char)(line[*pos] - 'a' + 10);
				else if('A' <= line[*pos] && line[*pos] <= 'F')
					num = (num << 4) + (unsigned char)(line[*pos] - 'A' + 10);
			}
			if(value)
				*value = num;
			if(n > 2)
				return 0;
			else
				return 1;
		}
		/* octal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; OCTAL(line[*pos]); *pos += 1, n++) {
			if(n == 2 && num > 037)
	       		n = 4;
			num = (num << 3) + (unsigned char)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 3)
			return 0;
		else
			return 1;
	}
	if(DECIMAL(line[*pos])) {
		/* decimal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; DECIMAL(line[*pos]); *pos += 1, n++) {
			if(n == 2 && num > 25)
	       		n = 4;
			if(n == 2 && num == 25 && line[*pos] > '5')
	       		n = 4;
			num = (num * 10) + (unsigned char)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 3)
			return 0;
		else
			return 1;
	}
	return 0;
}

static int ascii2integer8(char *line, int *pos, char *value)
{
	char num = 0;
	int n = 0;

	if(value)
		*value = 0;
	if(!line || !pos)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '0') {
		num = (unsigned char)(line[*pos] - '0');
		*pos += 1;
		if(line[*pos] == 'x' || line[*pos] == 'X') {
			*pos += 1;
			/* hexadecimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; HEXADECIMAL(line[*pos]); *pos += 1, n++) {
				if(DECIMAL(line[*pos]))
					num = (num << 4) + (unsigned char)(line[*pos] - '0');
				else if('a' <= line[*pos] && line[*pos] <= 'f')
					num = (num << 4) + (unsigned char)(line[*pos] - 'a' + 10);
				else if('A' <= line[*pos] && line[*pos] <= 'F')
					num = (num << 4) + (unsigned char)(line[*pos] - 'A' + 10);
			}
			if(value)
				*value = num;
			if(n > 2)
				return 0;
			else
				return 1;
		}
		/* octal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; OCTAL(line[*pos]); *pos += 1, n++) {
			if(n == 2 && num > 037)
	       		n = 4;
			num = (num << 3) + (unsigned char)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 3)
			return 0;
		else
			return 1;
	}
	if(line[*pos] == '-') {
		*pos += 1;
		if(DECIMAL(line[*pos])) {
			/* negative decimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; DECIMAL(line[*pos]); *pos += 1, n++) {
				if(n == 2 && num > 12)
		       		n = 4;
				if(n == 2 && num == 12 && line[*pos] > '8')
		       		n = 4;
				num = (num * 10) + (unsigned char)(line[*pos] - '0');
			}
			if(value)
				*value = num * (-1);
			if(n > 3)
				return 0;
			else
				return 1;
		}
	}
	if(line[*pos] == '+') {
		*pos += 1;
	}
	if(DECIMAL(line[*pos])) {
		/* positive decimal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; DECIMAL(line[*pos]); *pos += 1, n++) {
			if(n == 2 && num > 12)
	       		n = 4;
			if(n == 2 && num == 12 && line[*pos] > '7')
	       		n = 4;
			num = (num * 10) + (unsigned char)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 3)
			return 0;
		else
			return 1;
	}
	return 0;
}

static int ascii2unsigned16(char *line, int *pos, unsigned short *value)
{
	unsigned short num = 0;
	int n = 0;

	if(value)
		*value = 0;
	if(!line || !pos)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '0') {
		num = (unsigned short)(line[*pos] - '0');
		*pos += 1;
		if(line[*pos] == 'x' || line[*pos] == 'X') {
			*pos += 1;
			/* hexadecimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; HEXADECIMAL(line[*pos]); *pos += 1, n++) {
				if(DECIMAL(line[*pos]))
					num = (num << 4) + (unsigned short)(line[*pos] - '0');
				else if('a' <= line[*pos] && line[*pos] <= 'f')
					num = (num << 4) + (unsigned short)(line[*pos] - 'a' + 10);
				else if('A' <= line[*pos] && line[*pos] <= 'F')
					num = (num << 4) + (unsigned short)(line[*pos] - 'A' + 10);
			}
			if(value)
				*value = num;
			if(n > 4)
				return 0;
			else
				return 1;
		}
		/* octal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; OCTAL(line[*pos]); *pos += 1, n++) {
			if(n == 5 && num > 017777)
	       		n = 7;
			num = (num << 3) + (unsigned short)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 6)
			return 0;
		else
			return 1;
	}
	if(DECIMAL(line[*pos])) {
		/* decimal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; DECIMAL(line[*pos]); *pos += 1, n++) {
			if(n == 4 && num > 6553)
	       		n = 6;
			if(n == 4 && num == 6553 && line[*pos] > '5')
	       		n = 6;
			num = (num * 10) + (unsigned short)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 5)
			return 0;
		else
			return 1;
	}
	return 0;
}

static int ascii2integer16(char *line, int *pos, short *value)
{
	short num = 0;
	int n = 0;

	if(value)
		*value = 0;
	if(!line || !pos)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '0') {
		num = (unsigned short)(line[*pos] - '0');
		*pos += 1;
		if(line[*pos] == 'x' || line[*pos] == 'X') {
			*pos += 1;
			/* hexadecimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; HEXADECIMAL(line[*pos]); *pos += 1, n++) {
				if(DECIMAL(line[*pos]))
					num = (num << 4) + (unsigned short)(line[*pos] - '0');
				else if('a' <= line[*pos] && line[*pos] <= 'f')
					num = (num << 4) + (unsigned short)(line[*pos] - 'a' + 10);
				else if('A' <= line[*pos] && line[*pos] <= 'F')
					num = (num << 4) + (unsigned short)(line[*pos] - 'A' + 10);
			}
			if(value)
				*value = num;
			if(n > 4)
				return 0;
			else
				return 1;
		}
		/* octal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; OCTAL(line[*pos]); *pos += 1, n++) {
			if(n == 5 && num > 017777)
	       		n = 7;
			num = (num << 3) + (unsigned short)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 6)
			return 0;
		else
			return 1;
	}
	if(line[*pos] == '-') {
		*pos += 1;
		if(DECIMAL(line[*pos])) {
			/* negative decimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; DECIMAL(line[*pos]); *pos += 1, n++) {
				if(n == 4 && num > 3276)
		       		n = 6;
				if(n == 4 && num == 3276 && line[*pos] > '8')
		       		n = 6;
				num = (num * 10) + (unsigned short)(line[*pos] - '0');
			}
			if(value)
				*value = num * (-1);
			if(n > 5)
				return 0;
			else
				return 1;
		}
	}
	if(line[*pos] == '+') {
		*pos += 1;
	}
	if(DECIMAL(line[*pos])) {
		/* positive decimal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; DECIMAL(line[*pos]); *pos += 1, n++) {
			if(n == 4 && num > 3276)
	       		n = 6;
			if(n == 4 && num == 3276 && line[*pos] > '7')
	       		n = 6;
			num = (num * 10) + (unsigned short)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 5)
			return 0;
		else
			return 1;
	}
	return 0;
}

static int ascii2unsigned32(char *line, int *pos, unsigned long *value)
{
	unsigned long num = 0;
	int n = 0;

	if(value)
		*value = 0;
	if(!line || !pos)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '0') {
		num = (unsigned long)(line[*pos] - '0');
		*pos += 1;
		if(line[*pos] == 'x' || line[*pos] == 'X') {
			*pos += 1;
			/* hexadecimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; HEXADECIMAL(line[*pos]); *pos += 1, n++) {
				if(DECIMAL(line[*pos]))
					num = (num << 4) + (unsigned long)(line[*pos] - '0');
				else if('a' <= line[*pos] && line[*pos] <= 'f')
					num = (num << 4) + (unsigned long)(line[*pos] - 'a' + 10);
				else if('A' <= line[*pos] && line[*pos] <= 'F')
					num = (num << 4) + (unsigned long)(line[*pos] - 'A' + 10);
			}
			if(value)
				*value = num;
			if(n > 8)
				return 0;
			else
				return 1;
		}
		/* octal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; OCTAL(line[*pos]); *pos += 1, n++) {
			if(n == 10 && num > 03777777777UL)
	       		n = 12;
			num = (num << 3) + (unsigned long)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 11)
			return 0;
		else
			return 1;
	}
	if(DECIMAL(line[*pos])) {
		/* decimal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; DECIMAL(line[*pos]); *pos += 1, n++) {
			if(n == 9 && num > 429496729UL)
	       		n = 11;
			if(n == 9 && num == 429496729UL && line[*pos] > '5')
	       		n = 11;
			num = (num * 10) + (unsigned long)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 10)
			return 0;
		else
			return 1;
	}
	return 0;
}

static int ascii2integer32(char *line, int *pos, long *value)
{
	long num = 0;
	int n = 0;

	if(value)
		*value = 0;
	if(!line || !pos)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '0') {
		num = (unsigned long)(line[*pos] - '0');
		*pos += 1;
		if(line[*pos] == 'x' || line[*pos] == 'X') {
			*pos += 1;
			/* hexadecimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; HEXADECIMAL(line[*pos]); *pos += 1, n++) {
				if(DECIMAL(line[*pos]))
					num = (num << 4) + (unsigned long)(line[*pos] - '0');
				else if('a' <= line[*pos] && line[*pos] <= 'f')
					num = (num << 4) + (unsigned long)(line[*pos] - 'a' + 10);
				else if('A' <= line[*pos] && line[*pos] <= 'F')
					num = (num << 4) + (unsigned long)(line[*pos] - 'A' + 10);
			}
			if(value)
				*value = num;
			if(n > 8)
				return 0;
			else
				return 1;
		}
		/* octal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; OCTAL(line[*pos]); *pos += 1, n++) {
			if(n == 10 && num > 03777777777L)
	       		n = 12;
			num = (num << 3) + (unsigned long)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 11)
			return 0;
		else
			return 1;
	}
	if(line[*pos] == '-') {
		*pos += 1;
		if(DECIMAL(line[*pos])) {
			/* positive decimal number */
			for(; '0' == line[*pos]; *pos += 1)
				;
			for(; DECIMAL(line[*pos]); *pos += 1, n++) {
				if(n == 9 && num > 214748364L)
		       		n = 11;
				if(n == 9 && num == 214748364L && line[*pos] > '8')
		       		n = 11;
				num = (num * 10) + (unsigned long)(line[*pos] - '0');
			}
			if(value)
				*value = num * (-1L);
			if(n > 10)
				return 0;
			else
				return 1;
		}
	}
	if(line[*pos] == '+') {
		*pos += 1;
	}
	if(DECIMAL(line[*pos])) {
		/* positive decimal number */
		for(; '0' == line[*pos]; *pos += 1)
			;
		for(; DECIMAL(line[*pos]); *pos += 1, n++) {
			if(n == 9 && num > 214748364L)
	       		n = 11;
			if(n == 9 && num == 214748364L && line[*pos] > '7')
	       		n = 11;
			num = (num * 10) + (unsigned long)(line[*pos] - '0');
		}
		if(value)
			*value = num;
		if(n > 10)
			return 0;
		else
			return 1;
	}
	return 0;
}

static int ascii2domain(char *line, int *pos, unsigned char *buffer, int *length, int nbyte)
{
	int n;

	if(!line || !pos || !buffer || !length)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	for(n = 0, *length = 0; BASE64(line[*pos]) /*|| line[*pos] == '='*/; *pos += 1, n++) {
		if(n == 4) {
			*length += base64_decode((unsigned char*)&line[*pos-4], 4, &buffer[*length], (*length + 3) < nbyte ? 3 : nbyte - *length);
			n = 0;
		}
	}
	if(n == 4) {
		*length += base64_decode((unsigned char*)&line[*pos-4], 4, &buffer[*length], (*length + 3) < nbyte ? 3 : nbyte - *length);
	}
	if(n == 3) {
		*length += base64_decode((unsigned char*)&line[*pos-3], 3, &buffer[*length], (*length + 3) < nbyte ? 3 : nbyte - *length);
	}
	if(n == 2) {
		*length += base64_decode((unsigned char*)&line[*pos-2], 2, &buffer[*length], (*length + 2) < nbyte ? 2 : nbyte - *length);
	}
	if(n == 1) {
		*length += base64_decode((unsigned char*)&line[*pos-1], 1, &buffer[*length], (*length + 1) < nbyte ? 1 : nbyte - *length);
	}
	for(; line[*pos] == '='; *pos += 1)
		;
	return 1;
}

static int ascii2string(char *line, int *pos, char *buffer, int *length, int nbyte)
{
	if(!line || !pos || !buffer || !length)
		return 0;
	for(; WHITESPACE(line[*pos]); *pos += 1)
		;
	if(line[*pos] == '\"') {
		for(*pos += 1, *length = 0; ' ' <=  line[*pos] /*&& line[*pos] <= '\177'*/ && (line[*pos] != '\"' || line[*pos+1] == '\"'); *pos += 1) {
			if(line[*pos] == '\"' && line[*pos+1] == '\"')
				*pos += 1;
			if(*length + 1 < nbyte) {
				buffer[*length] = line[*pos];
				*length += 1;
			}
		}
	}
	else {
		for(*length = 0; ' ' <  line[*pos] /*&& line[*pos] <= '\177'*/; *pos += 1) {
			if(*length + 1 < nbyte) {
				buffer[*length] = line[*pos];
				*length += 1;
			}
		}
	}
	if(*length < nbyte)
		buffer[*length] = '\0';
	return 1;
}

/*	--------------------------------------------------------------------------
 *	Uwe Vogt,  UV Software,  Steinaecker 28,  88048 Friedrichshafen,  Germany
 *	Fon: +49-7541-6041530, Fax. +49-1803-551809359, Cell fon: +49-170-3801903
 *	E-Mail: uwe.vogt@uv-software.de, Internet URL: http://www.uv-software.de/
 */
